#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Useful functions and classes for handling callables.

NB: those functions are unable to manage built-in methods (so called
    method-wrappers). Unless specified otherwise, they will raise a
    C{TypeError} whenever passed such a callable.

@var  UNBOUNDED: returned by L{max_args} for callable accepting an arbitrary
                 number of arguments. UNBOUNDED is comparable to any integer,
                 and of course always considered greater. Also, most arithmetic
                 operations involving UNBOUNDED will return UNBOUNDED (all
                 except substracting, and dividing/modulo by UNBOUNDED, which
                 are not defined).

"""

from types import FunctionType, MethodType

MethodWrapperType = type (list.__getitem__.__call__)

def get_func (a_callable):
    """
    Return the function executed when the given callable is called.

    NB: if C{a_callable} is a class, C{__init__} is returned.

    @param a_callable: a callable

    @return: the function executed by C{a_callable}
    @rtype:  C{function}
    """
    return __get_func_and_reclevel (a_callable)[0]

def signature (a_callable, include_func=False):
    """
    Return the signature of the given callable.

    The signature is a triple C{(min,max,kw)} where:
      - C{min} is the minimum number of arguments accepted by the callable
      - C{max} is the maximum number of arguments accepted by the callable,
        -1 if unbounded
      - C{kw} is a boolean indicating whether the callable accepts

    If include_func is true, a 4th element, the function returned by get_func
    for a_callable, is included.
    """
    func, dec = __get_func_and_reclevel (a_callable)
    code = func.func_code
    mi = __min (func, code, dec)
    ma = __max (code, dec)
    kw = (code.co_flags & 0x08) != 0
    if include_func: return (mi, ma, kw, func)
    else:            return (mi, ma, kw)

def min_args (a_callable):
    """
    Return the minimum number of arguments accepted by the given callable.

    @param a_callable: a callable

    @return: the minimum number of arguments accepted by C{a_callable}
    @rtype:  C{int}
    """
    func, dec =  __get_func_and_reclevel (a_callable)
    code = func.func_code
    return __min (func, code, dec)

def max_args (a_callable):
    """
    Return the maximum number of arguments accepted by the given callable.

    @param a_callable: a callable

    @return: the maximum number of arguments accepted by C{a_callable}
    @rtype:  C{int} or L{UNBOUNDED}
    """
    func, dec = __get_func_and_reclevel (a_callable)
    code = func.func_code
    return __max (code, dec)

def kw_args (a_callable):
    """
    Return whether the given callable accepts arbitrary keyword arguments.

    @param a_callable: a callable
    """
    func = get_func (a_callable)
    code = func.func_code
    return code.co_flags & 0x08

def nblines (obj):
    """
    Return the number of lines of the given function.

    @param obj: a function or code object
    """
    if isinstance (obj, FunctionType):
        obj = obj.func_code
    vals = obj.co_lnotab[1::2]
    return sum ( ord(i) for i in vals ) + 1


class CallableWrapper (object):
    """
    Base class for wrapping callables.

    This class can be used as a base class for wrapping callables (e.g. for
    writing decorators). It acts as a proxy for usual attributes
    (C{__name__}, C{__doc__}, C{__module__}, C{func_name}, C{func_doc}) if the
    wrapped callable has them. All the methods of this module behave well with
    instances of this class (i.e. they get information from the wrapped
    callable rather than the C{__call__} function itself).

    Note also that CallableWrappers implement a C{__get__} method to return a
    method object, just as functions do, I{even} if the wrapped callable is not
    a function and has not such method.
    """
    def __init__ (self, a_callable):
        self.__c = a_callable
        for i in ("__name__","__doc__","__module__","func_name","func_doc"):
            try:
                setattr (self, i, getattr (a_callable, i))
            except AttributeError:
                pass

    def __call__ (self, *args, **kw):
        return self.__c (*args, **kw)

    __get__ = MethodType

    def _get_wrapped (self):
        "see: L{wrapped}"
        return self.__c

    wrapped = property (_get_wrapped, 
        doc = "Return the callable wrapped by this CallableWrapper" )



UNBOUNDED = None

class _Unbounded (object):
    def __new__ (_):
        global UNBOUNDED
        if UNBOUNDED is None:
            UNBOUNDED = object.__new__ (_Unbounded)
        return UNBOUNDED
    def __str__ (self):
        return 'UNBOUNDED'
    def __repr__ (self):
        return 'UNBOUNDED'
    def __cmp__ (self, other):
        if self is other:
            return 0
        else:
            return 1
    def __add__ (self, _): return self
    def __sub__ (self, _): return self
    def __mul__ (self, _): return self
    def __div__ (self, _): return self
    def __mod__ (self, _): return self
    def __radd__ (self, _): return self
    def __rmul__ (self, _): return self
    
    pass

_Unbounded () # instantiate singleton



# ================ Private toolkit ================

def __get_func_and_reclevel (a_callable):
    """
    Return the function underlying this callable, and the number of recursions
    needed to get it.

    @see: L{get_func}
    """
    if isinstance (a_callable, FunctionType):
        return a_callable, 0
    elif isinstance (a_callable, MethodWrapperType) :
        raise TypeError, "%s is a method-wrapper" % a_callable.__name__
    elif isinstance (a_callable, CallableWrapper):
        return __get_func_and_reclevel (a_callable._CallableWrapper__c)
    elif isinstance (a_callable, MethodType):
        f,r = __get_func_and_reclevel (a_callable.im_func)
        return f, r+1
    elif isinstance (a_callable, type):
        return __get_func_and_reclevel (a_callable.__init__)
    else:
        return __get_func_and_reclevel (a_callable.__call__)

def __min (func, code, dec):
    return max (0, code.co_argcount - len (func.func_defaults or []) - dec)

def __max (code, dec):
    if code.co_flags & 0x04:
        return UNBOUNDED
    else:
        return code.co_argcount - dec
