#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Harold programming tools.

This package contains a number of subpackages providing different programming
tools: 
  - L{harold.prog.abstract}
  - L{harold.prog.interface}
  - L{harold.prog.namespace}

Most of the functionalities provided by these packages are built-in in other
programming languages, and are considered to help programmers make better
software. However, they are sometimes also considered as annoying features
slowing programmers by forcing them to work a certain way.

The goal of this package is to keep the best of both approaches. All the
functionalities of the packages above are grouped in the L{HaroldClass}
metaclass. This metaclass performs a number of controls and operations on its
instance classes to provide the intented functionalities, respecting the
following principles:

  - C{HaroldClass} will never stop the program by raising an exception, but
    rather issue a warning whenever a control is not satisfied

  - C{HaroldClass} will also issue a warning whenever it changes the behaviour
    of its instance classes

Therefore, there are different ways of using harold:

  - people wishing a high level of control can turn the warnings into
    exceptions in order to be sure not to miss any;

  - people wishing to prototype rapidly can keep an eye on the warning but run
    their program anyway;

  - people not caring about some controls can filter out the corresponding warnings;

  - finally, people concerned about the performance cost of the controls can
    totally disable them by setting L{harold.lightweight}; this means that
    python library developped with harold can be imported without any
    additional cost.


@todo: Decide whether the use of C{harold.lightweight} should be replaced by
       the use of C{__debug__}.

@see: L{harold.lightweight}
"""

import harold

if harold.lightweight:
    HaroldClass = type
else:
    from harold.prog.abstract.abstractable_class import AbstractableClass
    from harold.prog.interface.interface_class import InterfaceClass
    from harold.prog.namespace.nsaware_class import NsAwareClass

    #
    # NB: The order of the subclasses below is important (could you believe
    # that!): NsAwareClass should be left of InterfaceClass, because ambiguous
    # members are overriden by an instance of AmbiguousMember which does not
    # satisfy InterfaceClass.
    #
    # Changing the order issues some unexpected warnings about callable members
    # overridden by non-callables when no such overriding is visible in the
    # class declaration.
    #
    class HaroldClass (NsAwareClass, AbstractableClass, InterfaceClass):
        """
        A metaclass providing functionnalities of all harold programming tools.

        Using this metaclass makes adds to your class L{contractual
        interfaces<harold.prog.interface>}, L{abstract members
        awareness<harold.prog.abstract>} and L{namespace
        disambiguation<harold.prog.namespace>}.

        While these functionalities are usually considered to help programmers
        make better software, they may be considered as CPU-consuming annoyance

        @note: if X{harold.lightweight} is set, this metaclass equals to
               L{type}
        """
        def __init__ (cls, cls_name, cls_mro, cls_members):
            super (HaroldClass, cls).__init__ (
                cls_members, cls_mro, cls_members)

        # this helps make some warnings more explicit
        _InterfaceClass__warning_level = 3
        _NsAwareClass__warning_level = 1

class HaroldClassWithAutoProperties (HaroldClass):
    """
    A metaclass providing functionnalities of all harold programming tools,
    and automatically adding properties when the class has method named
    according to the following pattern:
      - C{_get_I{name}} : attribute getter
      - C{_set_I{name}} : attribute setter
      - C{_del_I{name}} : attribute destructor
      - C{_doc_I{name}} : attribute docstring
    """
    def __init__ (cls, cls_name, cls_mro, cls_members):
        props = {}
        for name, f in cls_members.iteritems():
            try:
                nb_args = (f.func_code.co_argcount
                           - len (f.func_defaults or ()))
            except AttributeError:
                continue
            propname = None
            if name.startswith("_get_"):
                if nb_args == 1:
                    propname=name[5:]
            if name.startswith("_set_"):
                if nb_args == 2:
                    propname=name[5:]
            if propname:
                props[propname] = 1

        for propname in props.iterkeys ():
            fget = getattr(cls, "_get_%s" % propname, None)
            fset = getattr(cls, "_set_%s" % propname, None)
            fdel = getattr(cls, "_del_%s" % propname, None)
            sdoc = getattr(cls, "_doc_%s" % propname, None)
            setattr(cls, propname, property(fget, fset, fdel, sdoc))

        HaroldClass.__init__ (cls, cls_name, cls_mro, cls_members)
