#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Abstract classes for python.

Rationale
=========

It is common in object oriented languages to define a class which provides
common methods to several subclasses, but is not supposed to be instanciated,
generally because it is missing some methods. Such a class is known as an
X{abstract class}.

Being weakly typed, Python does not provide any explicit construct for abstract
classes. Indeed, so called abstract classes can be instanciated and this will
not even be a problem as long as missing methods are not invoked.  The drawback
of this flexibility is that I{when} such missing method is invoked, an
exception is raised at a point of the program where it can be quite difficult
to find the faulty instanciation.

This package provides an infrastructure allowing to explicitely define abstract
(i.e. missing) members, which has three advantages, without losing Python
flexibility:
  
  - it makes explicit the members to be defined in subclasses,

  - access to missing members raise a specific kind of exception, different
    from the C{AttributeError} raise by actually missing members,

  - it allows to detect, at instanciation time, whether the class is abstract
    or not, and to issue a warning or exception if it is.

The use of Python's warnings allows developers to fine tune the behaviour of
this package when an abstract class is instanciated.

Abstract members
================

The package implements the notion of X{abstract member}, I{i.e.} a method or
property which is not implemented by the class and has to be implemented by a
subclass.

C{abstract.attr_r} and C{abstract.attr_rw} can be used in class declaration to
provide the semantics of an abstract readable (resp. readable and writable)
attribute.  They can be used as decorators of a dummy method named after the
attribute, whose docstring will be used as the docstring of the attribute (it
may seem strange at first sight to declare attributes as methods, but it gives
an homogeneous and nice look to abstract classes, and makes the setting of a
docstring more convenient).
 
C{abstract.method} can be used as a function decorator to provide the semantics
of an abstract method.
 
Any attemps to access an abstract attribute will raise
C{AbstractAttributeException}. Any attempt to invoke an abstract method will
raise an C{AbstractMethodInvocationException}.

See L{harold.prog.abstract.member}

Abstract members and the MRO
----------------------------

It must be noted that python's MRO (method resolution order) influences the way
abstract methods are to be used. Consider the following example::
 
    class AbstractClass (object):
        __metaclass__ = AbstractableClass
        @abstract.method
        def m (self, x): pass

    class ConcreteClass (object):
        def m (self, x): return x

    class CorrectInheritance (ConcreteClass, AbstractClass):
        pass

    class MisleadingInheritance (AbstractClass, ConcreteClass):
        pass

The class C{CorrectInheritance} is concrete, since the MRO resolves method C{m}
as its concrete definition in C{ConcreteClass}. On the other hand,
C{MisleadingInheritance} is abstract because according to the MRO,
C{AbstractMethod.m} overrides C{ConcreteMethod.m}.

As a rule of thumb, it is always a good idea to put concrete classes first and
abstract classes last in the superclasses list. When using C{AbstractableClass}
(see below), failing to do so will issue an
{AbstractOverridesConcreteWarning<harold.prog.abstract.exceptions.AbstractOverridesConcreteWarning>}.


Abstractable class
==================

This package also provides the
L{AbstractableClass<abstractable_class.AbstractableClass>} metaclass. Any class
of this metaclass has a readonly attribute C{abstract}, indicating if it
contains or inherits non-overridden abstract members or cycles. In this case,
it will also have read-only attributes C{abstract_members} and
C{abstract_cycles}.

Furthermore, instantiating an abstract C{AbstractableClass} will issue an
L{AbstractClassInstantiationWarning<harold.prog.abstract.exceptions.AbstractClassInstantiationWarning>},
unless it has a I{default implementation} (see.below).

Abstract cycles
---------------

The package introduces the notion of X{abstract cyle}. Let us take an example::
 
    class Circle:
      def radius (self):
          return self.diameter()/2
      def diameter (self):
          return self.perimeter()/math.pi
      def perimeter (self):
          return self.radius()*math.pi*2
 
That class has no abstract method, since all are implemented. However, it I{is}
abstract, since none of its method will work properly (they will all result in
a stack overflow) unless at least one of them is overridden. Those three
methods are all implemented, but they form an abstract cycle (or dependancy
cycle) which must be broken by changing one of the implementations.
 
Each abstract cycle is declared as a tuple of methods which form the cycle.
All abstract cycles are declared as a list in a class attribute named
C{__abstract_cycles__}. Abstract cycles will B{not} prevent the methods to be
called, but an C{AbstractableClass} containing abstract cycles will recognizes
itself as abstract, just as if it contained an abstract member.

 
Default implementation
----------------------

Abstractable classes have an attribute C{default_implementation}. If the class
is concrete (i.e. if its attribute C{abstract} is false), the attribute is
read-only and equals the class. It the class is abstract, the attribute is
read-write. Any concrete subclass of the abstract class can be set as its
X{default implementation}. Any subsequent attempt to instantiate the abstract
class will return an instance of the subclass.  This replaces elegantly the
I{factory} idiom encountered in other languages.

Finally, the L{harold.prog.abstract} package can automatically set the
C{default_implementation} of abstract classes whenever a concrete subclass is
defined, if the abstract class has no default implementation yet. This is
enabled by the L{autoset_default_implementation<abstractable_class.autoset_default_implementation>} function. This behaviour is
disabled by default: since its consequences might be quite unexpected for
someone unaware of this functionnality, it has to be I{explicitely} enabled.
   
See L{harold.prog.abstract.abstractable_class}

"""
import harold

from harold.prog.abstract.member import *
