#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

from harold.prog.abstract.exceptions import (
    AbstractClassInstantiationWarning,
    AbstractOverridesConcreteWarning,
    ImplementationException,
)
from harold.prog.abstract.member import _AbstractMember

import warnings

warnings.filterwarnings ('always', category=AbstractClassInstantiationWarning)

_autoset_default_implementation = False

def autoset_default_implementation (bool):
    global _autoset_default_implementation
    _autoset_default_implementation = bool

class AbstractableClass (type):
    """
    A metaclass with checks about abstract members.

    See L{harold.prog.abstract} for more detail.
    """

    @staticmethod
    def __new_for_abstract_classes (cls, *args, **kw):
        #print "=== __new_for_abstract_classes", cls.__name__
        if cls.__abstract:
            di = cls.__default_implementation
            if di is not None:
                #print "    default impl. is ", di.__name__
                return di.__new__ (di, *args, **kw)
            else:
                warnings.warn (
                    "Instanciating abstract class %s" % cls,
                    AbstractClassInstantiationWarning,
                    stacklevel=2,
                )
        return cls.__real_new (cls, *args, **kw)

    def __init__ (cls, cls_name, direct_superclasses, direct_members):
        super (AbstractableClass, cls).__init__ (
            cls_name, direct_superclasses, direct_members )
        rmro = cls.mro()[::-1]
        members = {}
        abstract_cycles = []
        for a in rmro:
            #print "=== considering", a
            adict = a.__dict__
            for name,new in adict.iteritems () :
                #print "=== found", members.has_key (name) and "overriding for" or "", name

                if name in members \
                and     isinstance (new, _AbstractMember) \
                and not isinstance (members.get(name), _AbstractMember):
                    warnings.warn (
                        a.__name__,
                        AbstractOverridesConcreteWarning,
                        stacklevel=2,
                    )
                members[name] = new
                for cycle in abstract_cycles:
                    if name in cycle:
                        #print "=== found", name, "removing cycle", cycle
                        abstract_cycles.remove (cycle)
            abstract_cycles.extend (adict.get('__abstract_cycles__', []))

        # dedide whether I am abstract or not
        abstract_members = [ k for k,v in members.iteritems ()
                               if isinstance (v, _AbstractMember) ]
        if len (abstract_members) == 0 and len (abstract_cycles) == 0:
            #print "=== %s is concrete" % cls_name
            cls.__abstract = False
            if _autoset_default_implementation:
                for c in rmro:
                    if not isinstance(c, AbstractableClass):
                        continue
                    if c.__abstract and c.__default_implementation is None:
                        c.__default_implementation = cls
        else:
            #print "=== %s is abstract %s" % (cls_name, abstract_members)
            cls.__abstract = True
            cls.__abstract_members = tuple (abstract_members)
            cls.__abstract_cycles = tuple (abstract_cycles)
            cls.__default_implementation = None
            if cls.__new__ is not cls.__new_for_abstract_classes:
                #print "=== %s sets its __new__ method" % cls_name
                cls.__real_new = staticmethod (cls.__new__)
                cls.__new__    = staticmethod (cls.__new_for_abstract_classes)

    def get_abstract (cls):
        return cls.__abstract
    
    def get_abstract_members (cls):
        if cls.__abstract:
            return cls.__abstract_members
        else:
            return ()
    
    def get_abstract_cycles (cls):
        return cls.__abstract_cycles

    def get_default_implementation (cls):
        if cls.__abstract:
            return cls.__default_implementation
        else:
            return cls

    def set_default_implementation (cls, a_class):
        if not cls.__abstract:
            raise AttributeError, '%s is not abstract' % cls
        if not isinstance (a_class, type):
            if a_class is None:
                m = '\nTo unset the default implementation,' \
                    ' use the "del" operator.'
            else:
                m = ''
            raise ImplementationException, "%s is not a class%s" % (a_class, m)
        if cls not in a_class.mro ():
            raise ImplementationException, \
                  "%s is not a subclass of %s" % (cls, a_class)
        if a_class.__abstract:
            raise ImplementationException, "%s is abstract" % a_class
        cls.__default_implementation = a_class

    def del_default_implementation (cls):
        cls.__default_implementation = None

    abstract = property (get_abstract)

    abstract_members = property (get_abstract_members)

    abstract_cycles = property (get_abstract_cycles)

    default_implementation = property (get_default_implementation,
                                       set_default_implementation,
                                       del_default_implementation)
                

