#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Contractual interfaces for Python.

Rationale
=========

Let us illustrate the rationale of this package with a short example::

  class C:
    def m (self, x):
      return "I'm a C and I like", x

  class D (C):
    def m (self, x, y):
      return "I'm a D and I like", x, "and", y

  def proc (a_c):
    print a_c.m ("bananas")

  c1 = C()
  c2 = D()

  proc (c1) # works fine
  proc (c2) # raises a TypeError

Procedure C{proc} expects instances of class C{C}, and invokes their method
C{m}. C{c1} and C{c2} are both instances of C{C} because C{D} is a subclass of
C{C}. However, the last line raises an exception because C{D} overrides C{m} in
a way which is incompatible with the definition of C{C.m}: it expects 2
arguments instead of 1.

Such problems are usually considered as design flaws in a program, because they
do not respect the semantics of inheritance. Python being a weakly typed
language, it does not try to detect those problems at compile time. This is an
advantage in many situations, making prototyping easier by allowing unfinished
programs to run.

However, the drawback is that detecting such problems at runtime can be tricky,
even with thoroughful tests. This package proposes an infrastructure allowing
to detect such problems as soon as possible (i.e. when the class is loaded).



Contractual interface class
===========================

This package also provides the metaclass
L{interface_class.InterfaceClass}, whose class instances force
their subclasses to respect the interface of their members (except all those
starting with '__'). More precisely:

  - if the member is a I{property}, the overriding properties accept al least
    the same operations (get, set, del);

  - if the member is a I{callable}, the overriding callable should accept at
    least the same range of parameter number; furthermore, if the original
    member has a C{*args} (resp. C{**kw}) parameter, the overriding member
    should also have one.

Failing to fullfill those requirements will issue an
L{OverridingWarning<harold.prog.interface.exceptions.OverridingWarning>}. For
strict control, those warnings can be turned into exception with the standard
Python warning framework.

See L{interface_class.InterfaceClass} for usage examples.

Forcing members to be public
----------------------------

In some cases, some members whose name starts with '__' are to be considered
public with regard to the interface. This can be achieved by decorating the
given member with the L{public} decorator.

For example, the C{__init__} method is usually not considered part of the
contractual interface (because it defines the interface of the class itself
rather than the interface of instances, which is what we are adressing here).

However, there are situations where the subclasses are expected to produce
instances in the same fashion as the superclass. In this case, the C{__init__}
method will be decorated with L{public}.

Warning/Disclaimer
==================

It should be noted that C{InterfaceClass} does not (can not) guarantee that 
overriding members respect the I{semantics} of the original method. It does not
even check that a property or callable will not throw an exception when used
in a way it is supposed to work.

This package is of course not intended to replace good coding, but to help in
early detecting coding mistakes while (hopefuly ;-P ) preserving much of python
ease of use.

@todo: I plan to add method decorators to specify the expected types for
  callable parameters, and to ensure that the parameter types in overriding
  callables match those in the original callable.

@group Decorators: public
"""

def public (f):
    f.__harold_public__ = 1
    return f
