#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

from harold.prog.interface.exceptions import (
    OverridingWarning,
    NotCheckedWarning,
)

from harold.prog._deftracker import DefinitionsTracker, public_direct_members
from harold.callabletools    import signature, UNBOUNDED

from   inspect   import getargspec
from   itertools import izip
import warnings



def _check_compatible (name, old, new, warning_level):
    if isinstance (old, property):
        if isinstance (new, property):
            _check_compatible_property (name, old, new, warning_level+1)
        else:
            if old.fdel is not None:
                warnings.warn ( "%s overriding must be deletable" % name,
                                OverridingWarning, warning_level )
    elif callable (old):
        if callable (new):
            _check_compatible_callable (name, old, new, warning_level+1)
        else:
            warnings.warn ( "%s overriding must be callable" % name,
                            OverridingWarning, warning_level )
        
def _check_compatible_property (name, old, new, warning_level):
    if old.fget and not new.fget:
        warnings.warn ( "%s overriding must be gettable" % name,
                        OverridingWarning, warning_level)
    if old.fset and not new.fset:
        warnings.warn ( "%s overriding must be settable" % name,
                        OverridingWarning, warning_level)
    if old.fdel and not new.fdel:
        warnings.warn ( "%s overriding must be deletable" % name,
                        OverridingWarning, warning_level)
    if new.fdel and not old.fdel:
        warnings.warn ( "%s overriding can not be deletable" % name,
                        OverridingWarning, warning_level)

def _check_compatible_callable (name, old, new, warning_level):
    try:
        old_minp, old_maxp, old_kw, old_f = signature (old, True)
    except TypeError:
        warnings.warn ( "overridden %s is built-in" % old.__name__,
                        NotCheckedWarning, warning_level )
        return
    try:
        new_minp, new_maxp, new_kw, new_f = signature (new, True)
    except TypeError:
        warnings.warn ( "overriding %s is built-in" % new.__name__,
                        NotCheckedWarning, warning_level )
        return

    if new_minp > old_minp:
        warnings.warn ( 
            "%s overriding requires too many arguments (%s)" % (name, new_minp),
            OverridingWarning,
            warning_level,
        )
        return
    if new_maxp < old_maxp:
        if old_maxp == UNBOUNDED:
            warnings.warn (
                "%s overriding should have an '*args' argument " % name,
                OverridingWarning,
                warning_level,
            )
        else:
            warnings.warn (
                "%s overriding accepts too few arguments (%s)" %
                    (name, new_maxp),
                OverridingWarning,
                warning_level,
            )
        return
    if old_kw and not new_kw:
        warnings.warn ( "%s overriding should have a '**kw' argument " % name,
                        OverridingWarning, warning_level)
        return
    
    # checking for nested structure
    old_st = getargspec (old_f)[0]
    new_st = getargspec (new_f)[0]
    # some arguments in one function may be counted in *args for the other one
    # so we have to stock to the shortest list, which may be any of them:
    limit = min ((len (old_st), len (new_st)))
    old_st = old_st[:limit]
    new_st = new_st[:limit]
    if not __check_structure (old_st, new_st):
        warnings.warn ( "%s overriding has incompatible nested args" % name,
                       OverridingWarning, warning_level)


def __check_structure (old, new):
    if not isinstance (new, list):
        # single argument supercedes nested
        return True
    if not isinstance (old, list):
        return False
    l = len (old)
    if len (new) != l:
        return False
    for o,n in izip  (old, new):
        if not __check_structure (o, n):
            return False
    return True
        


class InterfaceClass (DefinitionsTracker):
    """
    A class with contractual interface.

    The following python code illustrate the use of C{InterfaceClass}::

      class X (object):
          __metaclass__ = InterfaceClass
          p_ro = property (g)
          p_rw = property (g,s)
          def m (self, x, y=1, z=1)

      class O1 (X):
          p_ro = property (None,s) 
          ## illegal: p_ro should be readable

      class O2 (X):
          p_ro = property (g,s)
          ## legal: p_ro is readable

      class O3 (X):
          p_rw = property (g)
          ## illegal: p_rw should be settable

      class O4 (X):
          def m (self)
          ## illegal: m should accept 2,3 or 4 parameters

      class O5 (X):
          def m (self, a, b, c)
          ## illegal: m should accept 2 or 3 parameters

      class O6 (X):
          def m (self, a)
          ## illegal: m should accept 3 or 4 parameters

      class O7 (X):
          def m (self, *args, **kw)
          ## legal: m does accept 2, 3 or 4 parameters
    """

    # this variable is used to make the warnings messages more explicit,
    # it works fine whith 0 when InterfaceClass is used directly,
    # but should be overridden in its subclasses
    __warning_level = 0

    def __init__ (cls, cls_name, direct_superclasses, direct_members):
        super (InterfaceClass, cls).__init__ (
            cls_name, direct_members, direct_members )

        public = dict (public_direct_members (cls, include__=False))
        for name,defs in cls._definitions().iteritems():
            if name not in public and isinstance (defs[0], InterfaceClass):
                continue
        #for name, new in public:
        #    defs = cls._definitions(name)

            new = getattr (cls, name) # get it with __get__ if available
            for d in defs:
                old = getattr (d, name)
                if old is not new:
                    _check_compatible (name, old, new, 3 + cls.__warning_level)
                    # may raise an exception
