#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Define the class L{IntervalsTree}.

An intervals-tree is a mapping between intervals and arbitrary values.  The
value associated to an interval can be overriden for a sub-interval, and thus
recursively. However, it is an error to assign different values to overlapping
intervals.

"""

class IntervalsTree (object):
    """
    @see: module description
    """
    def __init__ (self, default, *args):
        l = []
        self.__default = default
        self.__sons    = l
        if args:
            l.append (tuple (args))

    def __str__ (self):
        return "%s %s" % (self.__default, self.__sons)

    def _get_default (self):
        return self.__default
    def _set_default (self, val):
        self.__default = val
    default = property (_get_default, _set_default)

    def __getitem__ (self, i):
        for si,sj,s in self.__sons:
            if si <= i <= sj:
                if isinstance (s, IntervalsTree):
                    return s[i]
                else:
                    return s
        return self._get_default()

    def __setitem__ (self, (i,j), v):
        sons = self.__sons
        todel = None
        toappend = None
        for n in xrange (len (sons)):
            si,sj,s = sons[n]
            if isinstance (s, IntervalsTree):
                sv = s.default
            else:
                sv = s
                s = None

            if (si,sj) == (i,j):
                if sv != v:
                    raise ConflictException, ((si,sj),sv)
            elif si <= i and j <= sj:
                # (i,j) is *strictly* included in (si,sj)
                if s is not None:
                    s[i,j] = v
                    # do that even if v == sv because there may be in s an
                    # interval (ssi,ssj) s.t. si <= ssi <= i <= j <= ssj <= sj
                    # for which the default value ssv != v
                elif sv == v:
                    pass
                else:
                    sons[n] = (si, sj, IntervalsTree (sv, i, j, v))
            elif i <= si and sj <= j:
                # (si,sj) is *strictly* included in (i,j)
                if todel is None:
                    todel = [n]
                    if sv != v:
                        toappend = IntervalsTree (v, si, sj, s or sv)
                else:
                    todel.append (n)
                    if sv != v:
                        if toappend is None:
                            toappend = IntervalsTree (v, si, sj, s or sv)
                        else:
                            toappend.__sons.append((si, sj, s or sv))
                continue # (i,j) may include *several* (si,sj)'s
            elif si <= i <= sj or si <= j <= sj:
                raise ConflictException, ((si,sj),sv)
            else:
                continue
            break
        else:
            if todel is not None:
                todel.reverse()
                for n in todel:
                    del sons[n]

            toappend = toappend or v
            if v != self._get_default():
                sons.append ((i, j, toappend))

class ConflictException (Exception):
    """
    Raised whenever set tries to assign a value on an interval overlapping a
    different value.
    """
    pass
