#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

import harold

from harold.prog.namespace.exceptions import (
    AmbiguousMemberException,
    AmbiguousMemberWarning,
    ConflictingNamespacesException,
    InvalidContextWarning,
    NoDefinitionInNamespaceException,
    UnrecognizedContextException,
)
from harold.prog.namespace._intervals_tree import (
    IntervalsTree,
    ConflictException,
)

from harold import callabletools

from inspect   import stack
from itertools import chain
from types     import ModuleType
from warnings  import warn
from weakref   import ref

def ambiguous (*args, **kw):
    """
    Set an L{AmbiguousMember} to each given name in a class.
    
    Usage: ambiguous (name1[, name2, ...] [,automate=True])

    To be used inside a class definition.

    @see: L{package documentation<harold.prog.namespace>}
    """
    cls       = kw.get ("cls",       None)
    warn_only = kw.get ("warn_only", False)
    automate  = kw.get ("automate",  False)

    if cls is None:
        frame = stack()[1][0]
        try:
            cls_name = frame.f_code.co_name
            cls_dict = frame.f_locals
        finally:
            del frame
        for n in args:
            cls_dict[n] = AmbiguousMember (n, cls_name, warn_only, automate)
    else:
        for n in args:
            setattr (cls, n, AmbiguousMember (n, cls, warn_only, automate))
        

_EMPTY_TREE = IntervalsTree(None)

class AmbiguousMember (object):
    """
    An implementation artifact to manage ambiguous members.

    Instances of this class are not supposed to be constructed directly, but
    rather by the L{ambiguous} function.
    """

    def __init__ (self, name, cls, warn_only, automate):
        self.__name__ = name
        self.__warn_only = warn_only
        self.__automate = automate
        self.__ns = {}

        if isinstance (cls, type):
            self.__finish_init (cls)
        else:
            self.__class__ = _TransitionnalAmbiguousMember
            self._cls_name = cls

    def __finish_init (self, cls):
        self.__cls = ref (cls)
        name = self.__name__
        defs = []
        inherited = []
        for ancestor in cls.__mro__[1:][::-1]:
            d = ancestor.__dict__.get (name)
            if d is not None:
                if isinstance (d, AmbiguousMember):
                    if isinstance (d, _TransitionnalAmbiguousMember):
                        d.__get__ (None, ancestor)
                    inherited.insert (0, d.__get__ (None, ancestor))
                else:
                    defs = [ancestor] + [
                        a for a in defs if not issubclass (ancestor, a)
                    ]
        if len (defs) <= 1 and len (inherited) == 0:
            if len (defs) == 0:
                fmt = "Member %s is not defined"
            else:
                fmt = "Member %s is not ambiguous"
            raise TypeError, fmt % name

        self.__defs = defs
        self.__inherited = inherited

        if self.__automate:
            for c in defs:
                self.set_namespace (c, c)
        del self.__automate # not needed anymore


    def __str__ (self):
        return '%s.%s' % (self.__cls().__name__, self.__name__)

    def __protest (self):
        if self.__warn_only:
            warn (str(self), AmbiguousMemberWarning, stacklevel=3)
        else:
            raise AmbiguousMemberException, str(self)

    def __add_ns (self, namespace, filename, first=None, last=None):
        if filename[0] == '<':
            warn (filename, InvalidContextWarning, stacklevel=3)
            return
        ns = self.__ns
        it = ns.get (filename)
        if first is None:
            if it is None:
                ns[filename] = IntervalsTree (namespace)
            elif it.default is None:
                it.default = namespace
            elif it.default is not namespace:
                raise ConflictingNamespacesException, (it.default, filename)
        else:
            if last is None: last = first
            if it is None:
                it = IntervalsTree (None)
                ns[filename] = it
            try:
                it[first,last] = namespace
            except ConflictException, err:
                (i,j),ns = err.args
                raise ConflictingNamespacesException, (ns, filename, i, j)

    def set_namespace (self, namespace, context=1):
        """
        Assign a given namespace (i.e. class from which the ambiguous member
        should take its definition) for the given context.

        The context can be a callable, a property (in which case all methods of
        the property will taken into account), a class (in which case all
        members of the class will be taken into account), a module or an
        integer value.

        If an integer value M{n}, the namespace is assigned to the M{n} next
        lines of code. If not given, the context defaults to value 1 (i.e. the
        next line of code).

        IMPORTANT: the context must not be in the <stdin> file, because line
        number are nor reliable there. Any attempt to set a namespace to such
        a context will issue a warning and have no other effect.

        @param namespace: a class defining the member
        @param context: a callable, property, class, module or int

        @see: L{namespace_dec}
        """
        if isinstance (context, ModuleType):
            filename = context.__file__
            if filename[-3:] != '.py': filename = filename[:-1]
            self.__add_ns (namespace, filename)
        elif isinstance (context, type):
            for n,m in context.__dict__.iteritems():
                if isinstance (m, int):
                    continue # int is a valid context, but we do not want that
                try:
                    self.set_namespace (namespace, m)
                except UnrecognizedContextException:
                    pass
        elif callable (context):
            try:
                c = callabletools.get_func (context).func_code
            except TypeError:
                raise UnrecognizedContextException, context
                return
            file  = c.co_filename
            first = c.co_firstlineno
            last  = first + callabletools.nblines (c) - 1
            self.__add_ns (namespace, file, first, last)
        elif isinstance (context, property):
            for f in (context.fget, context.fset, context.fdel):
                if f is not None:
                    self.set_namespace (namespace, f)
        elif isinstance (context, int):
            frame = stack()[1][0]
            try:
                file = frame.f_code.co_filename
                lineno = frame.f_lineno
            finally:
                del frame
            self.__add_ns (namespace, file, lineno+1, lineno+context)
        else:
            raise UnrecognizedContextException, context
        return True

    def namespace_dec (self, namespace):
        """
        A decorator for assigning a namespace to a function.

        Both following codes are equivalent::
        
            # version with decorator
            @ambiguous_member.namespace_dec (A_Class)
            def f(x,y,z):
                pass

            # equivalent version without the decorator
            def f(x,y,z):
                pass
            ambiguous_member.set_namespace (A_Class, f)

        @see: L{set_namespace}
        """
        def decorator (f):
            self.set_namespace (namespace, f)
            return f
        return decorator

    def __get_namespace (self):
        ns         = self.__ns
        inherited  = self.__inherited
        convert    = self.__convert_namespace
        really_get = self._get_namespace

        for l in stack ()[2:]:
            frame = l[0]
            try:
                filename = frame.f_code.co_filename
                lineno = frame.f_lineno
            finally:
                del frame
            
            r = really_get (filename, lineno, ns, inherited, convert)
            if r is not None:
                return r

    def __get_definition (self, ns):
        try:
            return ns.__dict__[self.__name__]
        except KeyError:
            raise NoDefinitionInNamespaceException, ns

    if harold.lightweight:
        def __convert_namespace (self, n):
            for cls in self.__defs:
                if issubclass (cls, n):
                    return cls
            return n

        def _get_namespace (
                self, filename, lineno,
                ns=None, inherited=None, convert=None,
        ):
            """
            Return the first matching namespace for the given context.
            """
            if ns        is None:  ns        = self.__ns
            if inherited is None:  inherited = self.__inherited
            if convert   is None:  convert   = self.__convert_namespace

            r = ns.get (filename, _EMPTY_TREE) [lineno]
            if r is not None:
                return r
            for i in inherited:
                r = i._get_namespace (filename, lineno)
                if r is not None:
                    return convert (r)
    else:
        def __convert_namespace (self, n):
            found = None
            for cls in self.__defs:
                if issubclass (cls, n):
                    if found is None:
                        found = cls
                    else:
                        # this namespace is ambiguous, so
                        return None
            if found is None:
                return n
            else:
                return found

        def _get_namespace (
                self, filename, lineno,
                ns=None, inherited=None, convert=None,
        ):
            """
            Find all matching namespaces for the given context and check that
            there is no further ambiguity.
            """
            if ns        is None:  ns        = self.__ns
            if inherited is None:  inherited = self.__inherited
            if convert   is None:  convert   = self.__convert_namespace

            rf = None

            r = ns.get (filename, _EMPTY_TREE) [lineno]
            if r is not None:
                rf = r
            for i in inherited:
                r = i._get_namespace (filename, lineno)
                if r is not None:
                    r = convert (r)
                    if r is not None:
                        if rf is None:
                            rf = r
                        else:
                            if rf != r:
                                raise ConflictingNamespacesException, \
                                    ((rf, r), filename, lineno)
            return rf

    def __get__ (self, obj, typ=None):
        if obj is None:
            return self
        ns = self.__get_namespace()
        if ns is None:
            self.__protest()
            if obj is not None:
                return getattr (super (obj.__class__, obj), self.__name__)
            else:
                return getattr (super (typ, typ), self.__name__)
        else:
            d = self.__get_definition (ns)
            cls = d.__class__
            try:
                g = cls.__get__
            except AttributeError:
                return d
            return g (d, obj, typ)

    def __set__ (self, obj, value):
        ns = self.__get_namespace()
        if ns is None:
            self.__protest()
            return setattr (super (obj.__class__, obj), self.__name__, value)
        else:
            return self.__get_definition (ns).__set__ (obj, value)
            # NB: we do not try to solve the problem if the attribute has no
            #     __set__ method: setting it would add an entry in the 
            #     instance's __dict__, which is not a good idea since the
            #     member is ambiguous

    def __delete__ (self, obj):
        ns = self.__get_namespace()
        if ns is None:
            self.__protest()
            return delattr (super (obj.__class__, obj), self.__name__)
        else:
            return self.__get_definition (ns).__delete__ (obj)
            # NB: we do not try to solve the problem if the attribute has no
            #     __delete__ method: cf. __set__ above


class _TransitionnalAmbiguousMember (AmbiguousMember):

    def __finish_init (self, obj, typ=None):
        if typ is None: typ = obj.__class__
        name = self._cls_name
        for cls in typ.__mro__:
            if cls.__name__ == name:
                del self._cls_name
                break
        else:
            raise TypeError, "No ancestor %s" % name
        self._AmbiguousMember__finish_init (cls)
        self.__class__ = AmbiguousMember

    def __get__ (self, obj, typ=None):
        self.__finish_init (obj, typ)
        return self.__get__ (obj, typ)

    def __set__ (self, obj, value):
        self.__finish_init (obj)
        return self.__set__ (obj, value)

    def __delete__ (self, obj):
        self.__finish_init (obj)
        return self.__delete__ (obj)
