#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

"""
Useful functions and classes for building unit tests.

@group Decorators: assertNotRaise assertRaises failIfRaises failUnlessRaises
"""

__extra_epydoc_fields__ = [
    ('synonym', 'Synonym', 'Synonyms'),
]

from unittest import *

def failUnlessRaises (exception_class):
    """
    A parametrized decorator for test methods in test cases.
    It checks that the decorated function raises the given exception class.

    Usage example::

      class MyTest (TestCase):
          @failUnlessRaises (AttributeError)
          def test_x_has_no_y (self):
              "test that x has no attribute named y"
              self.x.y

    @param exception_class: a subclass of L{Exception} or a tuple of such 
                            subclasses
    @synonym: L{assertRaises}
    @synonym: L{failUnlessRaises}

    """
    def decorator (f):
        def wrapped (self, *args, **kw):
            self.failUnlessRaises (exception_class, f, self, *args, **kw)
        wrapped.func_name = f.func_name
        wrapped.__doc__ = f.__doc__
        return wrapped
    return decorator

assertRaises = failUnlessRaises

def failIfRaises (exception_class):
    """
    A decorator maker for test methods in test cases.
    It checks that the decorated function does not raise the given exception
    class.

    Also known as C{assertNotRaise}.

    Usage example::

      class MyTest (TestCase):
          @failIfRaises (AttributeError)
          def test_x_has_y (self):
              "test that x has an attribute named y"
              self.x.y

    @param exception_class: a subclass of L{Exception} or a tuple of such 
                            subclasses
    @synonym: L{assertNotRaise}
    @synonym: L{failIfRaises}
    """
    def decorator (f):
        def wrapped (self, *args, **kw):
            try:
                f (self, *args, **kw)
            except exception_class, msg:
                self.fail (msg)
        wrapped.func_name = f.func_name
        wrapped.__doc__ = f.__doc__
        return wrapped
    return decorator

assertNotRaise = failIfRaises
