#!/usr/bin/python2.4

#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

import harold
import harold.prog.abstract                    as     abstract
from   harold.prog.abstract.abstractable_class import *
from   harold.prog.abstract.exceptions         import *

from   harold.utesttools import failUnlessRaises, failIfRaises

import unittest
import warnings

if harold.lightweight:
    metaclass_to_test = type
else:
    metaclass_to_test = AbstractableClass

# ================ Abstract ================

class _TestAbstractBase (unittest.TestCase):
    def setUp (self):
        warnings.filterwarnings (
            'ignore',
            category=AbstractOverridesConcreteWarning,
        )
        # a class with abstract attributes
        class CA (object):
            __metaclass__ = metaclass_to_test

            @abstract.attr_r
            def ar ():
                pass

            @abstract.attr_rw
            def arw ():
                pass
        self.CA = CA
        # a class with abstract methods
        class CM (object):
            __metaclass__ = metaclass_to_test

            @abstract.method
            def m (self, a, b=1):
                pass

            @abstract.method
            def n (self, a, b=1):
                pass
        self.CM = CM
        # a class with abstract __init__
        class CI (object):
            __metaclass__ = metaclass_to_test

            @abstract.method
            def __init__ (self):
                pass
        self.CI = CI
        # a class with an abstract cycle
        class CC (object):
            __metaclass__ = metaclass_to_test

            def m1 (self):
                return self.m2 ()
            def m2 (self):
                return self.m1 ()

            __abstract_cycles__ = (('m1', 'm2'),)
        self.CC = CC
        # a concrete subclass of CA
        class CAC (CA):
            ar = 1
            arw = 1
        self.CAC = CAC
        # a concrete subclass of CM
        class CMC (CM):
            def m (self, a, b=1):
                return a,b 
            def n (self, a, b=1):
                return a,b
        self.CMC = CMC
        # a concrete subclass of CI 
        class CIC (CI):
            def __init__ (*args):
                return args
        self.CIC = CIC
        # a concrete subclass of CC
        class CCC (CC):
            def m2 (self):
                return 1
        self.CCC = CCC
    
    def tearDown (self):
        warnings.resetwarnings()


if not harold.lightweight:
    class TestMcAbstract (_TestAbstractBase):
        def testAbstractCA (self):
            "Class with abstract attribute is abstract"
            self.assert_ (self.CA.abstract)

        def testAbstractCM (self):
            "Class with abstract method is abstract"
            self.assert_ (self.CM.abstract)

        def testAbstractCI (self):
            "Class with abstract __init__ is abstract"
            self.assert_ (self.CI.abstract)
        
        def testAbstractCC (self):
            "Class with abstract cycle is abstract"
            self.assert_ (self.CC.abstract)

        def testAbstractSubclassCA (self):
            "Subclass of class with abstract attribute is abstract"
            class X (self.CA):
                ar = 1
            self.assert_ (X.abstract)

        def testAbstractSubclassCM (self):
            "Subclass of class with abstract method is abstract"
            class X (self.CM):
                def m (self, a, b=1):
                    return a,b
            self.assert_ (X.abstract)
        
        def testAbstractSubclassCI (self):
            "Subclass of class with abstract __init__ is abstract"
            class X (self.CM):
                def __init__ (self, a, b=1):
                    return a,b
            self.assert_ (X.abstract)
        
        def testAbstractSubclassCC (self):
            "Subclass of class with abstract cycle is abstract"
            class X (self.CC):
                pass
            self.assert_ (X.abstract)
        
        def testAbstractSubclassCC2 (self):
            "Subclass of class with more abstract cycles keeps all cycles"
            class X (self.CC):
                def m3 (self):
                    return self.m4 ()
                def m4 (self):
                    return self.m3 ()
                __abstract_cycles__ = (('m3', 'm4'),)
            self.assert_ (X.abstract)
            self.assertEquals (len (X.abstract_cycles), 2) #m1-2 and m3-4

        def testConcreteCAC (self):
            "Subclass defining abstract attribute is concrete"
            self.assert_ (not self.CAC.abstract)

        def testConcreteCMC (self):
            "Subclass defining abstract method is concrete"
            self.assert_ (not self.CMC.abstract)

        def testConcreteCIC (self):
            "Subclass defining abstract __init__ is concrete"
            self.assert_ (not self.CIC.abstract)
        
        def testConcreteCCC (self):
            "Subclass breaking abstract cycle is concrete"
            self.assert_ (not self.CCC.abstract)

class TestAbstractInstanciation (_TestAbstractBase):
    _fail = failUnlessRaises (AbstractClassInstantiationWarning)
    _pass = failIfRaises     (AbstractClassInstantiationWarning)

    def setUp (self):
        _TestAbstractBase.setUp (self)
        if not harold.lightweight:
            del self.CA.default_implementation
            del self.CM.default_implementation
        warnings.filterwarnings (
            'ignore',
            category=AbstractClassInstantiationWarning,
        )

    def tearDown (self):
        warnings.resetwarnings()

    @failUnlessRaises (AbstractAttributeException)
    def testGetAbstractAttribute (self):
        "Can not get abstract attribute"
        self.CA().ar
        self.CA().arw

    @failUnlessRaises (AttributeError)
    def testSetAbstractROAttribute (self):
        "Can not set abstract read-only attribute"
        self.CA().ar = 1

    @failUnlessRaises (AbstractAttributeException)
    def testSetAbstractRWAttribute (self):
        "Can not set abstract read-write attribute"
        self.CA().arw = 1

    @failUnlessRaises (AbstractMethodInvocationException)
    def testInvokeAbstractMethod (self):
        "Can not invoke abstract method"
        self.CM().m (1)
        self.CM().n (1)

    @failIfRaises (AbstractAttributeException)
    def testGetOverriddenAttribute (self):
        "Can get overriden attribute"
        self.CAC().ar
        self.CAC().arw

    @failIfRaises (AttributeError)
    def testSetOverriddenROAttribute (self):
        "Can set overriden read-only attribute"
        self.CAC().ar = 1
        # the RO attribute has ben overridden by a RW attribute

    @failIfRaises (AbstractAttributeException)
    def testSetOverriddenRWAttribute (self):
        "Can set overriden read-write attribute"
        self.CAC().arw = 1

    @failIfRaises (AbstractMethodInvocationException)
    def testInvokeOverriddenMethod (self):
        "Can invoke overriden method"
        self.CMC().m (1)
        self.CMC().n (1)


# ================ AbstractOverridesConcrete ================

class TestAbstractOverridesConcrete (unittest.TestCase):
    _pass = failIfRaises     (AbstractOverridesConcreteWarning)
    if harold.lightweight:
        _fail = _pass
    else:
        _fail = failUnlessRaises (AbstractOverridesConcreteWarning)

    def setUp (self):
        warnings.filterwarnings (
            'error',
            category=AbstractOverridesConcreteWarning,
        )
        class A (object):
            __metaclass__ = metaclass_to_test
            @abstract.method
            def m (self): pass
        self.A = A

        class C (object):
            def m (self): print "hello"
        self.C = C
    
    def tearDown (self):
        warnings.resetwarnings()

    @_pass
    def testCorrectOrder (self):
        class T (self.C, self.A):
            pass

    @_fail
    def testWrongOrder (self):
        class T (self.A, self.C):
            pass
        

# ================ Instantiation ================

class _TestInstantiationBase (unittest.TestCase):

    def setUp (self):
        warnings.filterwarnings (
            'error',
            category=AbstractClassInstantiationWarning,
        )
        if not harold.lightweight:
            autoset_default_implementation (True)

        # abstract class 1
        class A1 (object):
            __metaclass__ = metaclass_to_test

            @abstract.attr_r
            def ar ():
                pass
        self.A1 = A1
        # abstract class 2
        class A2 (object):
            __metaclass__ = metaclass_to_test

            @abstract.attr_r
            def br ():
                pass
        self.A2 = A2
        # concrete class
        class C1 (object):
            __metaclass__ = metaclass_to_test
            def __new__ (obj, *args, **kw):
                "This is a decoy"
                pass
        self.C1 = C1
        # concrete suclass of A2
        class A2C (A2):
            br = 1
        self.A2C = A2C

    def tearDown (self):
        warnings.resetwarnings()

_pass = failIfRaises     (AbstractClassInstantiationWarning)
if harold.lightweight:
    _fail = _pass
else:
    _fail = failUnlessRaises (AbstractClassInstantiationWarning)

class TestMcInstantiation (_TestInstantiationBase):

    @_fail
    def testInstantiateA1 (self):
        "Can not instantiate abstract class"
        self.A1 ()

    @_pass
    def testInstantiateC1 (self):
        "Can instantiate concrete class"
        self.C1 ()

    def testInstantiateA2 (self):
        "Can instantiate abstract cl. with default impl."
        if harold.lightweight:
            self.assert_ (isinstance (self.A2 (), self.A2))
        else:
            self.assert_ (isinstance (self.A2 (), self.A2C))

    @_fail
    def testSubclassA1Abstract (self):
        "Can not instantiate abstract subcl. of abstract cl."
        class X (self.A1):
            pass
        X ()

    @_pass
    def testSubclassA1Concrete (self):
        "Can instantiate concrete subcl. of abstract cl."
        class X (self.A1):
            ar = 1
        X ()

    @_fail
    def testSubclassA2Abstract (self):
        "Can not instantiate abstract subcl. of abstract cl. with default impl."
        class X (self.A2):
            pass
        X ()

    @_pass
    def testSubclassA2Concrete (self):
        "Can instantiate concrete subcl. of abstract cl. with default impl."
        class X (self.A2):
            br = 1
        X ()

    @_fail
    def testSubclassC1Abstract (self):
        "Can not instantiate abstract sub. of concrete cl."
        class X (self.C1):

            @abstract.attr_r
            def cr ():
                pass

        X ()

    @_pass
    def testSubclassC1Concrete (self):
        "Can instantiate concrete subcl. of concrete cl."
        class X (self.C1):
            pass
        X ()

    @_fail
    def testSubclassC1A1Abstract (self):
        "Can not instantiate subcl. of concrete and abstract"
        class X (self.C1, self.A1):
            pass
        X ()

    @_pass
    def testSubclassC1A1Concrete (self):
        "Can instantiate concrete subcl. of concrete and abstract"
        class X (self.C1, self.A1):
            ar = 1
        X ()

    @_fail
    def testSubclassC1A2A1Abstract (self):
        "Can not instantiate abstract subcl. of concrete and abstract with default impl."
        class X (self.C1, self.A2, self.A1):
            ar = 1
        X ()

    @_pass
    def testSubclassC1A2A1Concrete (self):
        "Can instantiate concrete subcl. of concrete and abstract with default impl."
        class X (self.C1, self.A2, self.A1):
            ar = 1
            br = 1
        X ()

# ================ Implementation ================

_fail = failUnlessRaises (ImplementationException)
_pass = failIfRaises     (ImplementationException)

if not harold.lightweight:
    class TestMcImplementation (_TestInstantiationBase):

        def testNoDefaultImplementationA1 (self):
            "Abstract class without any default implementation"
            self.assert_ (self.A1.default_implementation is None)

        def testHasDefaultImplementationA2 (self):
            "Abstract class with a default implementation"
            self.assert_ (self.A2.default_implementation is self.A2C)

        def testHasDefaultImplementationC1 (self):
            "Concrete class always has a default implementation (itself)"
            self.assert_ (self.C1.default_implementation is self.C1)

        def testHasDefaultImplementationA2C (self):
            "Concrete subclass of an abstract class is its own default impl."
            self.assert_ (self.A2C.default_implementation is self.A2C)

        @_fail
        def testSetDefaultImplementationA1None (self):
            "Default implementation can not be None"
            self.A1.default_implementation = None

        @_fail
        def testSetDefaultImplementationA1NotClass (self):
            "Default implementation must be a class"
            self.A1.default_implementation = 1

        @_fail
        def testSetDefaultImplementationA1NotSubclass (self):
            "Default implementation must be a subclass"
            self.A1.default_implementation = self.C1

        @_fail
        def testSetDefaultImplementationA1AbstractClass (self):
            "Default implementation must be concrete"
            class X (self.A1):
                pass
            self.A1.default_implementation = X

        @_pass
        def testSetDefaultImplementationA1Pass (self):
            "Setting default implementation"
            class X (self.A1):
                ar = 1
            self.A1.default_implementation = X
            self.assert_ (isinstance (self.A1 (), X))

        @_fail
        def testSetDefaultImplementationA2None (self):
            "New default implementation can not be None"
            self.A2.default_implementation = None

        @_fail
        def testSetDefaultImplementationA2NotClass (self):
            "New default implementation must be a class"
            self.A2.default_implementation = 1

        @_fail
        def testSetDefaultImplementationA2NotSubclass (self):
            "New default implementation must be a subclass"
            self.A2.default_implementation = self.C1

        @_fail
        def testSetDefaultImplementationA2AbstractClass (self):
            "New default implementation must be concrete"
            class X (self.A2):
                pass
            self.A2.default_implementation = X

        @_pass
        def testSetDefaultImplementationA2Pass (self):
            "Changing default implementation"
            class X (self.A2):
                br = 1
            self.A2.default_implementation = X
            self.assert_ (isinstance (self.A2 (), X))

        def testSetDefaultImplementationC1 (self):
            "Can not change default implementation of concrete class"
            class X (self.C1):
                pass
            def assign ():
                self.C1.default_implementation = X
            self.assertRaises (AttributeError, assign)



if __name__ == "__main__":
    unittest.main ()
