#!/usr/bin/python2.4

#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

import harold
import harold.prog.interface as interface
from   harold.prog.interface.interface_class import InterfaceClass
from   harold.prog.interface.exceptions import *
from   harold.utesttools import failUnlessRaises, failIfRaises

import unittest
import warnings

if harold.lightweight:
    metaclass_to_test = type
    _fail = failIfRaises     (OverridingWarning)
    _pass = failIfRaises     (OverridingWarning)
else:
    metaclass_to_test = InterfaceClass
    _fail = failUnlessRaises (OverridingWarning)
    _pass = failIfRaises     (OverridingWarning)

# ================ Common base class ================

class _TestInterface (unittest.TestCase):
    def setUp (self):
        warnings.filterwarnings ("error", category=OverridingWarning)
        class TestIC (object):
            __metaclass__ = metaclass_to_test
    
            pr = property (lambda x,y: 1)
            prw = property (lambda x,y: 1, lambda x,y:1)
            pw = property (None, lambda x,y: 1)
            prwd = property (*[lambda x,y: 1]*3)

            def __init__ (self, a, b=1): pass
            def m0 (self): pass
            def m01 (self, a=1): pass
            def m02 (self, a=1, b=1): pass
            def m03 (self, a=1, b=1, c=1): pass
            def m0p (self, *args): pass
            def m0pk (self, *args, **kw): pass
            def m0k (self, **kw): pass
            def m1 (self, a): pass
            def m12 (self, a, b=1): pass
            def m13 (self, a, b=1, c=1): pass
            def m1p (self, a, *args): pass
            def m1pk (self, a, *args, **kw): pass
            def m1k (self, a, **kw): pass
            def m2 (self, a, b): pass
            def m23 (self, a, b, c=1): pass
            def m2p (self, a, b, *args): pass
            def m2pk (self, a, b, *args, **kw): pass
            def m2k (self, a, b, **kw): pass
            def m3 (self, a, b, c): pass
            def m3p (self, a, b, c, *args): pass
            def m3pk (self, a, b, c, *args, **kw): pass
            def m3k (self, a, b, c, **kw): pass

            def __private (self, a, b, c, d, e): pass

        self.TestIC = TestIC
        self.base_classes = (TestIC,)

    def tearDown (self):
        warnings.resetwarnings()

# ================ Test case overriding nothing ================

class TestMcOverridingNone (_TestInterface):

    @_pass
    def testOverridingNone (self):
        "Subclassing but defining other members"
        class X (self.TestIC):
            q = property (lambda s: 1)
            def n (self, a,b,c,d,e):
                pass

# ================ Test case overriding attributes ================

class TestMcOverridingAttributes (_TestInterface):

    @_pass 
    def testPrAtom (self):
        "Overridding read-only attribute with a value"
        class X (self.TestIC):
            pr = 1

    @_pass
    def testPrPropertyG (self):
        "Overridding read-only attribute with a read-only property"
        class X (self.TestIC):
            pr = property (lambda x,y: 1)

    @_pass
    def testPrPropertyGS (self):
        "Overridding read-only attribute with a read-write property"
        class X (self.TestIC):
            pr = property (lambda x,y: 1, lambda x,y: 1)

    @_fail
    def testPrPropertyS (self):
        "Overridding read-only attribute with a write-only property"
        class X (self.TestIC):
            pr = property (None, lambda x,y: 1)

    @_fail
    def testPrPropertyD (self):
        "Overridding read-only attribute with a delete-only property"
        class X (self.TestIC):
            pr = property (None, None, lambda x: 1)
   
    @_fail
    def testPrPropertyGSD (self):
        "Overridding read-only attribute with a full-fledged property"
        class X (self.TestIC):
            pr = property (*[lambda x: 1]*3)


    @_pass
    def testPrwAtom (self):
        "Overridding read-write attribute with a value"
        class X (self.TestIC):
            prw = 1

    @_fail
    def testPrwPropertyG (self):
        "Overridding read-write attribute with a read-only property"
        class X (self.TestIC):
            prw = property (lambda x,y: 1)

    @_pass
    def testPrwPropertyGS (self):
        "Overridding read-write attribute with a read-write property"
        class X (self.TestIC):
            prw = property (lambda x,y: 1, lambda x,y: 1)

    @_fail
    def testPrwPropertyS (self):
        "Overridding read-write attribute with a write-only property"
        class X (self.TestIC):
            prw = property (None, lambda x,y: 1)

    @_fail
    def testPrwPropertyD (self):
        "Overridding read-write attribute with a delete-only property"
        class X (self.TestIC):
            prw = property (None, None, lambda x: 1)
   
    @_fail
    def testPrwPropertyGSD (self):
        "Overridding read-write attribute with a full-fledged property"
        class X (self.TestIC):
            prw = property (*[lambda x: 1]*3)


    @_pass
    def testPwAtom (self):
        "Overridding write-only attribute with a value"
        class X (self.TestIC):
            pw = 1

    @_fail
    def testPwPropertyG (self):
        "Overridding write-only attribute with a read-only property"
        class X (self.TestIC):
            pw = property (lambda x,y: 1)

    @_pass
    def testPwPropertyGS (self):
        "Overridding write-only attribute with a read-write property"
        class X (self.TestIC):
            pw = property (lambda x,y: 1, lambda x,y: 1)

    @_pass
    def testPwPropertyS (self):
        "Overridding write-only attribute with a write-only property"
        class X (self.TestIC):
            pw = property (None, lambda x,y: 1)

    @_fail
    def testPwPropertyD (self):
        "Overridding write-only attribute with a delete-only property"
        class X (self.TestIC):
            pw = property (None, None, lambda x: 1)
   
    @_fail
    def testPwPropertyGSD (self):
        "Overridding write-only attribute with a full-fledged property"
        class X (self.TestIC):
            pw = property (*[lambda x: 1]*3)


    @_fail
    def testPrwdAtom (self):
        "Overridding full-fledged attribute with a value property"
        class X (self.TestIC):
            prwd = 1

    @_fail
    def testPrwdPropertyG (self):
        "Overridding full-fledged attribute with a read-only property"
        class X (self.TestIC):
            prwd = property (lambda x,y: 1)

    @_fail
    def testPrwdPropertyGS (self):
        "Overridding full-fledged attribute with a read-write property"
        class X (self.TestIC):
            prwd = property (lambda x,y: 1, lambda x,y: 1)

    @_fail
    def testPrwdPropertyS (self):
        "Overridding full-fledged attribute with a write-only property"
        class X (self.TestIC):
            prwd = property (None, lambda x,y: 1)

    @_fail
    def testPrwdPropertyD (self):
        "Overridding full-fledged attribute with a delete-only property"
        class X (self.TestIC):
            prwd = property (None, None, lambda x: 1)
   
    @_pass
    def testPrwdPropertyGSD (self):
        "Overridding full-fledged attribute with a full-fledged property"
        class X (self.TestIC):
            prwd = property (*[lambda x: 1]*3)

# ================ Test cases overriding methods ================

class _OverridingBase (object):
    """
    A method store for testing the overriding of methods in InterfaceClass.

    It is assumed to be inherited in conjunction with L{_TestInterface}, which
    provide C{self.TestIC}. It does not inherit it directly because we do not
    want this class to inherit C{unittest.TestCase}, to prevent
    C{unittest.main} from using it.

    TestCase inheriting this class must:
      - inherit L{_TestInterface} as well
      - provide a C{method_name} attribute: the method of C{self.TestIC} to be 
        overridden
      - provide a C{method_desc} attribute to appear in the TestCase short
        descriptions
      - provide an attribute C{pass_testI{X}} for each C{testI{X}} which is
        B{not} supposed to raise a
        L{harold.prog.interface.exceptions.OverridingWarning}; 
        otherwise, all methods are supposed to fail with this exception.

    The test methods are named as follow:
      - the minimum number of parameters
      - if different, the maximum number of parameters
      - if the method has C{*args}, the letter 'p' (stands for 'plus')
      - if the method has C{**kw}, the letter 'k'

    Note also that those method actually perform 3 tests in one:
      - overriding with a function
      - overriding with a callable
      - overriding with a class
    """

    def _make_test (f):
        """
        Decorator to make the actual test, and make it aware of the
        corresponding C{fail_X} constant.
        """
        def test_w_func (self, *args, **kw):
            metaclass_to_test ('X', self.base_classes, {self.method_name: f})

        def test_w_callable (self, *args, **kw):
            class TheCallable (object):
                __call__ = f
            c = TheCallable()
            metaclass_to_test ('X', self.base_classes, {self.method_name: c})

        def test_w_class (self, *args, **kw):
            class c (object):
                __init__ = f
            metaclass_to_test ('X', self.base_classes, {self.method_name: c})

        def wrapper (self, *args, **kw):
                
            must_pass = hasattr (self, 'pass_%s' % f.func_name)
            if must_pass or harold.lightweight:
                try:
                    test_w_func     (self, *args, **kw)
                    test_w_callable (self, *args, **kw)
                    test_w_class    (self, *args, **kw)
                except OverridingWarning, msg:
                    self.fail (msg)
            else:
                self.failUnlessRaises (
                    OverridingWarning,
                    test_w_func,
                    self,
                    *args,
                    **kw
                )
        wrapper.__name__ = f.__name__
        wrapper.__doc__  = f.__doc__
        return wrapper

    def shortDescription (self):
        sd = unittest.TestCase.shortDescription (self)
        return sd % self.method_desc


    @_make_test
    def test0 (self):
        "Overriding %s with 0 argument"
        pass

    @_make_test
    def test01 (self, x=0):
        "Overriding %s with 0-1 argument"
        pass

    @_make_test
    def test02 (self, x=0, y=0):
        "Overriding %s with 0-2 arguments"
        pass
    
    @_make_test
    def test03 (self, x=0, y=0, z=0):
        "Overriding %s with 0-3 arguments"
        pass

    @_make_test
    def test04 (self, x=0, y=0, z=0, w=0):
        "Overriding %s with 0-4 arguments"
        pass

    @_make_test
    def test0p (self, *args):
        "Overriding %s with 0+ arguments"
        pass

    @_make_test
    def test0pk (self, *args, **kw):
        "Overriding %s with 0+ arguments & keywords"
        pass

    @_make_test
    def test0k (self, **kw):
        "Overriding %s with keywords"
        pass
    
    @_make_test
    def test1 (self, x):
        "Overriding %s with 1 argument"
        pass

    @_make_test
    def test12 (self, x, y=0):
        "Overriding %s with 1-2 arguments"
        pass

    @_make_test
    def test13 (self, x, y=0, z=0):
        "Overriding %s with 1-3 arguments"
        pass

    @_make_test
    def test14 (self, x, y=0, z=0, w=0):
        "Overriding %s with 1-4 arguments"
        pass

    @_make_test
    def test1p (self, x, *args):
        "Overriding %s with 1+ argument"
        pass

    @_make_test
    def test1pk (self, x, *args, **kw):
        "Overriding %s with 1+ argument & keywords"
        pass

    @_make_test
    def test1k (self, x, **kw):
        "Overriding %s with 1 argument & keywords"
        pass

    @_make_test
    def test2 (self, x, y):
        "Overriding %s with 2 arguments"
        pass
   
    @_make_test
    def test23 (self, x, y, z=0):
        "Overriding %s with 2-3 arguments"
        pass

    @_make_test
    def test24 (self, x, y, z=0, w=0):
        "Overriding %s with 2-4 arguments"
        pass

    @_make_test
    def test2p (self, x, y, *args):
        "Overriding %s with 2+ arguments"
        pass

    @_make_test
    def test2pk (self, x, y, *args, **kw):
        "Overriding %s with 2+ arguments & keywords"
        pass

    @_make_test
    def test2k (self, x, y, **kw):
        "Overriding %s with 2 arguments & keywords"
        pass

    @_make_test
    def test3 (self, x, y, z):
        "Overriding %s with 3 arguments"
        pass

    @_make_test
    def test34 (self, x, y, z, w=0):
        "Overriding %s with 3-4 arguments"
        pass

    @_make_test
    def test3p (self, x, y, z, *args):
        "Overriding %s with 3+ arguments"
        pass

    @_make_test
    def test3pk (self, x, y, z, *args, **kw):
        "Overriding %s with 3+ arguments & keywords"
        pass

    @_make_test
    def test3k (self, x, y, z, **kw):
        "Overriding %s with 3 arguments & keywords"
        pass

    @_make_test
    def test4 (self, x, y, z, w):
        "Overriding %s with 4 arguments"
        pass

    @_make_test
    def test4p (self, x, y, z, w, *args):
        "Overriding %s with 4+ arguments"
        pass

    @_make_test
    def test4pk (self, x, y, z, w, *args, **kw):
        "Overriding %s with 4+ arguments & keywords"
        pass

    @_make_test
    def test4k (self, x, y, z, w, **kw):
        "Overriding %s with 4 arguments & keywords"
        pass


class TestMcOverridingMethod0 (_OverridingBase, _TestInterface):
    method_name = "m0"
    method_desc = "method (0)"

    pass_test0 = 1
    pass_test01 = 1
    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test0k = 1

class TestMcOverridingMethod01 (_OverridingBase, _TestInterface):
    method_name = "m01"
    method_desc = "method (0-1)"

    pass_test01 = 1
    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1

class TestMcOverridingMethod02 (_OverridingBase, _TestInterface):
    method_name = "m02"
    method_desc = "method (0-2)"

    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1

class TestMcOverridingMethod03 (_OverridingBase, _TestInterface):
    method_name = "m03"
    method_desc = "method (0-3)"

    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1

class TestMcOverridingMethod0p (_OverridingBase, _TestInterface):
    method_name = "m0p"
    method_desc = "method (0+)"

    pass_test0p = 1
    pass_test0pk = 1

class TestMcOverridingMethod0pk (_OverridingBase, _TestInterface):
    method_name = "m0pk"
    method_desc = "method (0+ & kw)"

    pass_test0pk = 1

class TestMcOverridingMethod0pk (_OverridingBase, _TestInterface):
    method_name = "m0k"
    method_desc = "method (0 & kw)"

    pass_test0k = 1
    pass_test0pk = 1

class TestMcOverridingMethod1 (_OverridingBase, _TestInterface):
    method_name = "m1"
    method_desc = "method (1)"

    pass_test01 = 1
    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test1 = 1
    pass_test12 = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test1k = 1

class TestMcOverridingMethod12 (_OverridingBase, _TestInterface):
    method_name = "m12"
    method_desc = "method (1-2)"

    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test12 = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1

class TestMcOverridingMethod13 (_OverridingBase, _TestInterface):
    method_name = "m13"
    method_desc = "method (1-3)"

    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1

class TestMcOverridingMethod1p (_OverridingBase, _TestInterface):
    method_name = "m1p"
    method_desc = "method (1+)"

    pass_test0p = 1
    pass_test0pk = 1
    pass_test1p = 1
    pass_test1pk = 1

class TestMcOverridingMethod1pk (_OverridingBase, _TestInterface):
    method_name = "m1pk"
    method_desc = "method (1+ & kw)"

    pass_test0pk = 1
    pass_test1pk = 1

class TestMcOverridingMethod1k (_OverridingBase, _TestInterface):
    method_name = "m1k"
    method_desc = "method (1 & kw)"

    pass_test0pk = 1
    pass_test1pk = 1
    pass_test1k = 1

class TestMcOverridingMethod2 (_OverridingBase, _TestInterface):
    method_name = "m2"
    method_desc = "method (2)"

    pass_test02 = 1
    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test12 = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test2 = 1
    pass_test23 = 1
    pass_test24 = 1
    pass_test2p = 1
    pass_test2pk = 1
    pass_test2k = 1

class TestMcOverridingMethod23 (_OverridingBase, _TestInterface):
    method_name = "m23"
    method_desc = "method (2-3)"

    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test23 = 1
    pass_test24 = 1
    pass_test2p = 1
    pass_test2pk = 1

class TestMcOverridingMethod2p (_OverridingBase, _TestInterface):
    method_name = "m2p"
    method_desc = "method (2+)"

    pass_test0p = 1
    pass_test0pk = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test2p = 1
    pass_test2pk = 1

class TestMcOverridingMethod2pk (_OverridingBase, _TestInterface):
    method_name = "m2pk"
    method_desc = "method (2+ & kw)"

    pass_test0pk = 1
    pass_test1pk = 1
    pass_test2pk = 1

class TestMcOverridingMethod2k (_OverridingBase, _TestInterface):
    method_name = "m2k"
    method_desc = "method (2 & kw)"

    pass_test0pk = 1
    pass_test1pk = 1
    pass_test2k = 1
    pass_test2pk = 1

class TestMcOverridingMethod3 (_OverridingBase, _TestInterface):
    method_name = "m3"
    method_desc = "method (3)"

    pass_test03 = 1
    pass_test04 = 1
    pass_test0p = 1
    pass_test0pk = 1
    pass_test13 = 1
    pass_test14 = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test23 = 1
    pass_test24 = 1
    pass_test2p = 1
    pass_test2pk = 1
    pass_test3 = 1
    pass_test34 = 1
    pass_test3p = 1
    pass_test3pk = 1
    pass_test3k = 1

class TestMcOverridingMethod3p (_OverridingBase, _TestInterface):
    method_name = "m3p"
    method_desc = "method (3+)"

    pass_test0p = 1
    pass_test0pk = 1
    pass_test1p = 1
    pass_test1pk = 1
    pass_test2p = 1
    pass_test2pk = 1
    pass_test3p = 1
    pass_test3pk = 1

class TestMcOverridingMethod3pk (_OverridingBase, _TestInterface):
    method_name = "m3pk"
    method_desc = "method (3+ & kw)"

    pass_test0pk = 1
    pass_test1pk = 1
    pass_test2pk = 1
    pass_test3pk = 1

class TestMcOverridingMethod3k (_OverridingBase, _TestInterface):
    method_name = "m3k"
    method_desc = "method (3 & kw)"

    pass_test0pk = 1
    pass_test1pk = 1
    pass_test2pk = 1
    pass_test3pk = 1
    pass_test3k = 1

class TestMcOverridingPrivateMethod (_OverridingBase, _TestInterface):
    method_name = "__private"
    method_desc = "private method"

    pass_test0 = pass_test01 = pass_test02 = pass_test03 = pass_test04 = 1
    pass_test1 = pass_test12 = pass_test13 = pass_test14 = 1
    pass_test2 = pass_test23 = pass_test24 = 1
    pass_test3 = pass_test34 = 1
    pass_test4 = 1
    pass_test0p = pass_test0pk = pass_test0k = 1
    pass_test1p = pass_test1pk = pass_test1k = 1
    pass_test2p = pass_test2pk = pass_test2k = 1
    pass_test3p = pass_test3pk = pass_test3k = 1
    pass_test4p = pass_test4pk = pass_test4k = 1

class TestMcOverridingInitPrivate (TestMcOverridingPrivateMethod):
    method_name = "__init__"
    method_desc = "__init__ (private)"
    # pass_X inherited from TestMcOverridingPrivateMethod

class TestMcOverridingInitPublic (TestMcOverridingMethod12):
    method_name = "__init__"
    method_desc = "__init__ (public)"
    # pass_X inherited from TestMcOverridingMethod12

    def setUp (self):
        TestMcOverridingMethod12.setUp (self)
        self.TestIC.__init__ = interface.public (self.TestIC.__init__.im_func)


# =============== Test cases overriding several methods at once ===============

class _TestInterfaceSeveral (_TestInterface):
    def setUp (self):
        super (_TestInterfaceSeveral, self).setUp()

        def m1 (self, a, **kw):
            pass
        def m12 (self, a, b, c=1):
            pass
        def m1k (self, a):
            pass
        def m2 (self, a, b, c):
            pass
        def m23 (self, a, b=1):
            pass
        def m3 (self, a, b):
            pass

        TestIC2 = metaclass_to_test ('TestIC2', (object,), {
                self.method_name : locals()[self.method_name],
            }
        )
        self.base_classes = (self.TestIC, TestIC2)

class TestMcOverridingMethod1_1k (_TestInterfaceSeveral, TestMcOverridingMethod1k):
    method_name = "m1"
    method_desc = "methods (1|1 & kw)"
    # pass_X inherited from TestMcOverridingMethod1k

class TestMcOverridingMethod12_23 (_TestInterfaceSeveral, TestMcOverridingMethod13):
    method_name = "m12"
    method_desc = "methods (1-2|2-3)"
    # pass_X inherited from TestMcOverridingMethod13

class TestMcOverridingMethod1K_1 (_TestInterfaceSeveral, TestMcOverridingMethod1k):
    method_name = "m1k"
    method_desc = "methods (1 & kw|1)"
    # pass_X inherited from TestMcOverridingMethod1k

class TestMcOverridingMethod2_3 (_TestInterfaceSeveral, TestMcOverridingMethod23):
    method_name = "m2"
    method_desc = "methods (2|3)"
    # pass_X inherited from TestMcOverridingMethod23

class TestMcOverridingMethod23_12 (_TestInterfaceSeveral, TestMcOverridingMethod13):
    method_name = "m23"
    method_desc = "methods (2-3|1-2)"
    # pass_X inherited from TestMcOverridingMethod13

class TestMcOverridingMethod3_2 (_TestInterfaceSeveral, TestMcOverridingMethod23):
    method_name = "m3"
    method_desc = "methods (3|2)"
    # pass_X inherited from TestMcOverridingMethod23


# =============== Test methods with nested arguments ===============

class TestMcNestedArguments (unittest.TestCase):
    def setUp (self):
        class C(object):
            __metaclass__ = metaclass_to_test
            def m (self, a, (b, c, d)): pass
        self.C = C
        warnings.filterwarnings ("error", category=OverridingWarning)

    def tearDown (self):
        warnings.resetwarnings()

    @_pass
    def testNotNested (self):
        class D(self.C):
            def m (self, a, z): pass

    @_fail
    def testTooShort (self):
        class D(self.C):
            def m (self, a, (b, c)): pass

    @_pass
    def testRightSize (self):
        class D(self.C):
            def m (self, a, (b, c, d)): pass

    @_fail
    def testTooLong (self):
        class D(self.C):
            def m (self, a, (b, c, d, e)): pass
    pass
            
# =============== Test methods uncheckable types ===============

if harold.lightweight:
    _fail = failIfRaises     (NotCheckedWarning)
    _pass = failIfRaises     (NotCheckedWarning)
else:
    _fail = failUnlessRaises (NotCheckedWarning)
    _pass = failIfRaises     (NotCheckedWarning)

class TestMcNotChecked (unittest.TestCase):
    def setUp (self):
        warnings.filterwarnings ("ignore", category=NotCheckedWarning)
        class C(object):
            __metaclass__ = metaclass_to_test

        class D(dict,C):
            pass

        class E(C):
            def items (self): pass

        self.C = C
        self.D = D
        self.E = E

        warnings.resetwarnings()
        warnings.filterwarnings ("error", category=NotCheckedWarning)

    def tearDown (self):
        warnings.resetwarnings()

    @_fail
    def testOldIsUncheckable (self):
        class FOld(self.C, dict):
            def items (self): pass

    @_fail
    def testOldIsUncheckable2 (self):
        class FOld2(self.D):
            def items (self): pass

    @_fail
    def testNewIsUncheckable (self):
        class FNew(dict, self.E):
            pass


# =============== main ===============

if __name__ == "__main__":
    unittest.main ()
