#!/usr/bin/python2.4

#
# (c)2005 LIRIS - University Claude Bernard Lyon 1
# http://liris.cnrs.fr/
#
# Author: Pierre-Antoine CHAMPIN
# http://champin.net/
#
# This software is distributed under the terms of the GNU LGPL v2.1.
# See LICENSE.txt for more details.
#

import harold

import unittest

from inspect import stack
from warnings import filterwarnings, resetwarnings

from harold.callabletools import get_func
from harold.prog.namespace import ambiguous
from harold.prog.namespace.exceptions import *
from harold.prog.interface.exceptions import OverridingWarning

from   harold.utesttools import failUnlessRaises, failIfRaises

from _namespace_module import *
import _namespace_module


if harold.lightweight:
    metaclass_to_test   = type
    _fail_with_acw      = failIfRaises (AmbiguousClassWarning)
    _pass_without_acw   = failIfRaises (AmbiguousClassWarning)
    _fail_with_amw      = failIfRaises (AmbiguousMemberWarning)
    _pass_without_amw   = failIfRaises (AmbiguousMemberWarning)
else:
    from harold.prog.namespace.nsaware_class import NsAwareClass
    metaclass_to_test   = NsAwareClass
    _fail_with_acw      = failUnlessRaises (AmbiguousClassWarning)
    _pass_without_acw   = failIfRaises     (AmbiguousClassWarning)
    _fail_with_amw      = failUnlessRaises (AmbiguousMemberWarning)
    _pass_without_amw   = failIfRaises     (AmbiguousMemberWarning)


MODULE = _namespace_module.__file__
if MODULE[-1] != 'y': MODULE = MODULE[:-1]

HERE = __file__
if HERE[-1] != 'y': HERE = HERE[:-1]


class Musician (object):
    def play (self, instrument):
        return "Playing the %s" % instrument
    L_playM = LINENO()-1
    def do_piano (self):
        return self.play ("piano")
    L_do_piano = LINENO()-1
    def get_instrument (self):
        return getattr (self, "_instrument", "piano")
    L_get_instrument = LINENO()-1
    def set_instrument (self, value):
        self._instrument = value
    def del_instrument (self):
        try: del self._instrument
        except AttributeError: pass
    hobby = property (get_instrument, set_instrument, del_instrument)

class LousyMusician (Musician):
    def play (self, instrument):
        return "%s lousily" % super (LousyMusician, self).play (instrument)

class Player (object):
    def play (self, game):
        return "Playing %s" % game
    L_playP = LINENO()-1
    def do_go (self):
        return self.play ("go")
    L_do_go = LINENO()-1
    def get_game (self):
        return getattr (self, "_game", "go")
    L_get_game = LINENO()-1
    def set_game (self, value):
        self._game = value
    def del_game (self):
        try: del self._game
        except AttributeError: pass
    hobby = property (get_game, set_game, del_game)

class Kid (object):
    def play (self, toy):
        return "Playing with my %s" % toy
    def do_go (self):
        return self.play ("teddy")
    def get_toy (self):
        return getattr (self, "_toy", "teddy")
    def set_toy (self, value):
        self._toy = value
    def del_toy (self):
        try: del self._toy
        except AttributeError: pass
    hobby = property (get_toy, set_toy, del_toy)


# ================ TestCases ================

class _TestBase (object):
    def setUp (self):
        class MP (Musician, Player):
            ambiguous ("play", "hobby")
        self.MP = MP
        

class TestGetNamespace (_TestBase, unittest.TestCase):
    def setUp (self):
        super (TestGetNamespace, self).setUp()
        self.lineno1 = SomeClass.L_get_attr
        self.lineno2 = SomeClass.L_call

    def testNone (self):
        m = self.MP.play
        self.assertEquals (m._get_namespace (HERE, LINENO()), None)

    def testContext (self):
        m = self.MP.play

        m.set_namespace (Musician)
        self.assertEquals (m._get_namespace (HERE, LINENO()), Musician)
        self.assertEquals (m._get_namespace (HERE, LINENO()), None)

        m.set_namespace (Musician, 2)
        pass
        self.assertEquals (m._get_namespace (HERE, LINENO()), Musician)
        self.assertEquals (m._get_namespace (HERE, LINENO()), None)

    def testFunction (self):
        m = self.MP.play
        m.set_namespace (Musician, some_func)
        self.assertEquals (m._get_namespace (MODULE, L_some_func), Musician)
        self.assertEquals (m._get_namespace (MODULE, L_some_func + 1), None)

    def testMethod (self):
        m = self.MP.play
        m.set_namespace (Musician, SomeClass.get_attr)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1 + 1), None)

    def testProperty (self):
        m = self.MP.play
        m.set_namespace (Musician, SomeClass.attr)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1 + 1), None)

    def testClass (self):
        m = self.MP.play
        m.set_namespace (Musician, SomeClass)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2 + 1), None)

    def testModule (self):
        m = self.MP.play
        m.set_namespace (Musician, _namespace_module)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2), Musician)
        self.assertEquals (m._get_namespace (HERE, LINENO()), None)

    def testCallable (self):
        m = self.MP.play
        c = SomeClass (self.MP())
        m.set_namespace (Musician, c)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), None)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2 + 1), None)

    def testOverride (self):
        m = self.MP.play
        m.set_namespace (Musician, SomeClass.get_attr)
        m.set_namespace (Player, _namespace_module)
        self.assertEquals (m._get_namespace (MODULE, self.lineno1), Musician)
        self.assertEquals (m._get_namespace (MODULE, self.lineno2), Player)
        self.assertEquals (m._get_namespace (HERE, LINENO()), None)

class TestUseNamespace (_TestBase, unittest.TestCase):
    def setUp (self):
        super (TestUseNamespace, self).setUp()
        self.m = self.MP()
        self.o = SomeClass (self.m)

    def testWrongGet (self):
        def get_hobby (x): return x.hobby
        self.assertRaises (AmbiguousMemberException, get_hobby, self.m)
    
    def testWrongSet (self):
        def set_hobby (x): x.hobby = "foo"
        self.assertRaises (AmbiguousMemberException, set_hobby, self.m)

    def testWrongDel (self):
        def del_hobby (x): del x.hobby
        self.assertRaises (AmbiguousMemberException, del_hobby, self.m)

    def testGetDefault (self):
        self.MP.hobby.set_namespace (Musician)
        self.assertEqual (self.m.hobby, "piano")
        self.MP.hobby.set_namespace (Player)
        self.assertEqual (self.m.hobby, "go")

    def testGet (self):
        self.m._instrument = "trumpet"
        self.MP.hobby.set_namespace (Musician)
        self.assertEqual (self.m.hobby, "trumpet")

        self.m._game = "cards"
        self.MP.hobby.set_namespace (Player)
        self.assertEqual (self.m.hobby, "cards")

    def testSet (self):
        self.MP.hobby.set_namespace (Musician)
        self.m.hobby = "trumpet"
        self.assertEqual (self.m._instrument, "trumpet")

        self.MP.hobby.set_namespace (Player)
        self.m.hobby = "cards"
        self.assertEqual (self.m._game, "cards")

    def testDel (self):
        self.m._instrument = "trumpet"
        self.MP.hobby.set_namespace (Musician, 2)
        del self.m.hobby
        self.assertEqual (self.m.hobby, "piano")

        self.m._game = "cards"
        self.MP.hobby.set_namespace (Player, 2)
        del self.m.hobby
        self.assertEqual (self.m.hobby, "go")

    def testCallMusician (self):
        MP = self.MP
        MP.play.set_namespace (Musician, 2)
        try:
            self.assertEqual (MP().play("trumpet"), "Playing the trumpet")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)

    def testCallPlayer (self):
        MP = self.MP
        MP.play.set_namespace (Player, 2)
        try:
            self.assertEqual (MP().play("cards"), "Playing cards")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)

    def testDownInStackMusician (self):
        MP = self.MP
        o = self.o
        MP.play.set_namespace (Musician, some_func)
        try:
            self.assertEqual (o(), "Playing the foo")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)

    def testDownInStackPlayer (self):
        MP = self.MP
        o = self.o
        MP.play.set_namespace (Player, some_func)
        try:
            self.assertEqual (o(), "Playing foo")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)

    def testUpInStackMusician (self):
        MP = self.MP
        o = self.o
        MP.play.set_namespace (Musician, o)
        try:
            self.assertEqual (o(), "Playing the foo")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)

    def testUpInStackPlayer (self):
        MP = self.MP
        o = self.o
        MP.play.set_namespace (Player, o)
        try:
            self.assertEqual (o(), "Playing foo")
        except AmbiguousMemberException, err:
            self.fail ("Member is still ambiguous %s" % err)


class TestInheritance1 (TestGetNamespace):
    def setUp (self):
        super (TestInheritance1, self).setUp ()
        MP = self.MP
        class LMP1 (LousyMusician, MP):
            ambiguous ("play", "hobby")
        self.MP = LMP1

class TestInheritance2 (TestGetNamespace):
    def setUp (self):
        super (TestInheritance2, self).setUp ()
        MP = self.MP
        class LMP2 (MP, LousyMusician):
            ambiguous ("play", "hobby")
        self.MP = LMP2

class TestInconsistentInheritance (_TestBase, unittest.TestCase):
    def setUp (self):
        super (TestInconsistentInheritance, self).setUp ()
        MP = self.MP
        class MK (Musician, Kid):
            ambiguous ("play", "hobby")
        self.MK = MK
        class MKP (MP, MK):
            ambiguous ("play", "hobby")
        self.MKP = MKP

    def testInconsistent (self):
        lineno = SomeClass.L_get_attr
        self.MP.play.set_namespace (Musician, SomeClass.get_attr)
        self.MK.play.set_namespace (Kid, SomeClass.get_attr)
        o = SomeClass (self.MKP())
        if harold.lightweight:
            try:
                o.get_attr()
            except ConflictingNamespacesException:
                self.fail ("Unexpected conflict in lightweight mode")
        else:
            self.assertRaises (ConflictingNamespacesException, o.get_attr)

class TestAutomated (unittest.TestCase):
    def setUp (self):
        super (TestAutomated, self).setUp ()
        class MP (Musician, Player):
            ambiguous ("play", "hobby", automate=1)
        self.MP = MP

    def testAutomated (self):
        m = self.MP.play
        h = self.MP.play
        M = Musician
        P = Player
        self.assertEquals (m._get_namespace (HERE, M.L_playM), Musician)
        self.assertEquals (m._get_namespace (HERE, M.L_do_piano), Musician)
        self.assertEquals (m._get_namespace (HERE, P.L_playP), Player)
        self.assertEquals (m._get_namespace (HERE, P.L_do_go), Player)
        self.assertEquals (h._get_namespace (HERE, M.L_playM), Musician)
        self.assertEquals (h._get_namespace (HERE, M.L_do_piano), Musician)
        self.assertEquals (h._get_namespace (HERE, P.L_playP), Player)
        self.assertEquals (h._get_namespace (HERE, P.L_do_go), Player)


class TestMc (unittest.TestCase):
    def setUp (self):
        class MP (Musician, Player):
            pass
        self.MP = MP
        filterwarnings ("error", category=AmbiguousClassWarning)
        filterwarnings ("ignore", category=OverridingWarning)

    def tearDown (self):
        resetwarnings()

    @_fail_with_acw
    def testAmbiguousDeclaration (self):
        class MP2 (self.MP):
            __metaclass__ = metaclass_to_test

    @_pass_without_acw
    def testKnowinglyAmbiguousDeclaration (self):
        class MP2 (self.MP):
            __metaclass__ = metaclass_to_test
            ambiguous ("play", "hobby")
    
    @_pass_without_acw
    def testDisambiguatedDeclaration (self):
        class MP2 (self.MP):
            __metaclass__ = metaclass_to_test
            def play (self, stuff):
                pass
            hobby = Musician.hobby 
 
class TestMcUse (unittest.TestCase):
    def setUp (self):
        filterwarnings ("ignore", category=AmbiguousClassWarning)
        filterwarnings ("error", category=AmbiguousMemberWarning)
        class MP (Musician, Player):
            __metaclass__ = metaclass_to_test
        self.mp = MP()

    def tearDown (self):
        resetwarnings()

    @_pass_without_amw
    def testUnambiguousMember (self):
        self.mp.get_instrument()

    @_fail_with_amw
    def testAmbiguousMember (self):
        self.mp.play ("piano")

    @_fail_with_amw
    def testIndirectAmbiguousMember (self):
        some_func (self.mp)

    @_pass_without_amw
    def testAutomatedDisambiguation (self):
        self.mp.do_piano ()
    

class TestOverride (unittest.TestCase):
    def setUp (self):
        class C (object):
            __metaclass__ = metaclass_to_test
            def m (self): pass
        self.C = C
        class D (object):
            def n (self): pass
        self.D = D
        class E (object):
            def m (self): pass
            def n (self): pass
        self.E = E
        filterwarnings ("error", category=AmbiguousClassWarning)

    def tearDown (self):
        resetwarnings()

    @_fail_with_acw
    def testWithoutOverrideSingle (self):
        class F (self.E, self.C):
            pass

    @_pass_without_acw
    def testWithOverrideSingle (self):
        class F (self.E, self.C):
            __override__ = { self.E: self.C }

    @_fail_with_acw
    def testWithoutOverrideMultiple (self):
        class F (self.E, self.D, self.C):
            __override__ = { self.E: self.C }

    @_pass_without_acw
    def testWithOverrideMultiple (self):
        class F (self.E, self.D, self.C):
            __override__ = { self.E: (self.D, self.C) }



if __name__ == "__main__":
    unittest.main ()
